# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import typing
from builtins import *

import canmatrix


class CanCluster(dict):

    def __init__(self, *arg, **kw):
        super(CanCluster, self).__init__(*arg, **kw)
        self._frames = []  # type: typing.List[canmatrix.Frame]
        self._signals = []  # type: typing.List[canmatrix.Signal]
        self._ecus = []  # type: typing.List[canmatrix.Ecu]
        self._pdu_gateway_list = []     # type: typing.List[dict[str, str]]
        self._signal_gateway_list = []  # type: typing.List[dict[str, str]]
        self.update()

    def update_frames(self):  # type: () -> typing.MutableSequence[canmatrix.Frame]
        frames = []  # type: typing.List[canmatrix.Frame]
        frame_names = []  # type: typing.List[str]
        for matrixName in self:
            for frame in self[matrixName].frames:  # type: canmatrix.Frame
                if frame.name not in frame_names:
                    frame_names.append(frame.name)
                    frames.append(frame)
                else:
                    index = frame_names.index(frame.name)
                    for transmitter in frame.transmitters:
                        frames[index].add_transmitter(transmitter)
                    for receiver in frame.receivers:
                        frames[index].add_receiver(receiver)
        self._frames = frames
        return frames

    def update_signals(self):  # type: () -> typing.MutableSequence[canmatrix.Signal]
        signals = []  # type: typing.List[canmatrix.Signal]
        signal_names = []  # type: typing.List[str]
        for matrixName in self:
            for frame in self[matrixName].frames:  # type: canmatrix.Frame
                for signal in frame.signals:
                    if signal.name not in signal_names:
                        signal_names.append(signal.name)
                        signals.append(signal)
                    else:
                        index = signal_names.index(signal.name)
                        for receiver in signal.receivers:
                            signals[index].add_receiver(receiver)
        self._signals = signals
        return signals

    def update_ecus(self):  # type: () -> typing.MutableSequence[canmatrix.Ecu]
        ecus = []  # type: typing.List[canmatrix.Ecu]
        ecu_names = []  # type: typing.List[str]
        for matrixName in self:
            for ecu in self[matrixName].ecus:  # type: canmatrix.Ecu
                if ecu.name not in ecu_names:
                    ecu_names.append(ecu.name)
                    ecus.append(ecu)
        self._ecus = ecus
        return ecus

    def update(self):
        self.update_frames()
        self.update_signals()
        self.update_ecus()

    @property
    def ecus(self):  # type: () -> typing.MutableSequence[canmatrix.Ecu]
        if not self._ecus:
            self.update_ecus()
        return self._ecus

    @property
    def frames(self):  # type: () -> typing.MutableSequence[canmatrix.Frame]
        if not self._frames:
            self.update_frames()
        return self._frames

    @property
    def signals(self):  # type: () -> typing.MutableSequence[canmatrix.Signal]
        if not self._signals:
            self.update_signals()
        return self._signals

    def pdu_gateway(self, pdu_gateway_list=[]):
        self._pdu_gateway_list += pdu_gateway_list
        return self._pdu_gateway_list

    def signal_gateway(self, signal_gateway_list=[]):
        self._signal_gateway_list += signal_gateway_list
        return self._signal_gateway_list

    def get_pdu_routing_info(self, pdu_name, strict_search=False):
        routing_source = []
        routing_target = []
        if strict_search:
            for pdu in self._pdu_gateway_list:
                if pdu_name == pdu["source"]:
                    routing_source.append({"pdu": pdu["target"], "cluster": pdu["target_cluster"], "ecu": pdu["ecu"], "type": pdu["target_type"]})
                if pdu_name == pdu["target"]:
                    routing_target.append({"pdu": pdu["source"], "cluster": pdu["source_cluster"], "ecu": pdu["ecu"], "type": pdu["source_type"]})
        else:
            for pdu in self._pdu_gateway_list:
                if pdu_name in pdu["source"]:
                    routing_source.append({"pdu": pdu["target"], "cluster": pdu["target_cluster"], "ecu": pdu["ecu"], "type": pdu["target_type"]})
                if pdu_name in pdu["target"]:
                    routing_target.append({"pdu": pdu["source"], "cluster": pdu["source_cluster"], "ecu": pdu["ecu"], "type": pdu["source_type"]})
        return {"source": routing_source, "target": routing_target}

    def get_signal_routing_info(self, signal_name, strict_search=False):
        routing_source = []
        routing_target = []
        if strict_search:
            for signal_gw in self._signal_gateway_list:
                if signal_name == signal_gw["source"]:
                    routing_source.append({"signal": signal_gw["target"], "cluster": signal_gw["target_cluster"], "ecu": signal_gw["ecu"], "type": signal_gw["target_type"]})
                if signal_name == signal_gw["target"]:
                    routing_target.append({"signal": signal_gw["source"], "cluster": signal_gw["source_cluster"], "ecu": signal_gw["ecu"], "type": signal_gw["source_type"]})
        else:
            for signal_gw in self._signal_gateway_list:
                if signal_name in signal_gw["source"]:
                    routing_source.append({"signal": signal_gw["target"], "cluster": signal_gw["target_cluster"], "ecu": signal_gw["ecu"], "type": signal_gw["target_type"]})
                if signal_name in signal_gw["target"]:
                    routing_target.append({"signal": signal_gw["source"], "cluster": signal_gw["source_cluster"], "ecu": signal_gw["ecu"], "type": signal_gw["source_type"]})
        return {"source": routing_source, "target": routing_target}



# -*- coding: utf-8 -*-
# Copyright (c) 2013, Eduard Broecker
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that
# the following conditions are met:
#
#    Redistributions of source code must retain the above copyright notice, this list of conditions and the
#    following disclaimer.
#    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

#
# this script exports json-files from a canmatrix-object
# json-files are the can-matrix-definitions of the CANard-project
# (https://github.com/ericevenchick/CANard)

from __future__ import absolute_import, division, print_function

import json
import sys
import typing
from builtins import *
import decimal
import canmatrix


def dump(db, f, **options):
    # type: (canmatrix.CanMatrix, typing.BinaryIO, **str) -> None

    export_canard = options.get('jsonExportCanard', False)
    motorola_bit_format = options.get('jsonMotorolaBitFormat', "lsb")
    export_all = options.get('jsonExportAll', False)
    native_types = options.get('jsonNativeTypes', False)
    number_converter = float if native_types else str
    additional_frame_columns = [x for x in options.get("additionalFrameAttributes", "").split(",") if x]


    export_dict = {}
    if export_all:
        export_dict['enumerations'] = db.value_tables

    export_dict['messages'] = []

    if export_canard:
        for frame in db.frames:
            signals = {}
            for signal in frame.signals:
                signals[
                    signal.get_startbit(
                        bit_numbering=1,
                        start_little=True)] = {
                    "name": signal.name,
                    "bit_length": signal.size,
                    "factor": signal.factor,
                    "offset": signal.offset}
            export_dict['messages'].append(
                {"name": frame.name, "id": hex(frame.arbitration_id.id), "signals": signals})

    elif export_all is False:
        for frame in db.frames:
            symbolic_signals = []
            for signal in frame.signals:
                if not signal.is_little_endian:
                    if motorola_bit_format == "msb":
                        start_bit = signal.get_startbit(bit_numbering=1)
                    elif motorola_bit_format == "msbreverse":
                        start_bit = signal.get_startbit()
                    else:  # motorola_bit_format == "lsb"
                        start_bit = signal.get_startbit(bit_numbering=1, start_little=True)
                else:
                    start_bit = signal.get_startbit(bit_numbering=1, start_little=True)

                symbolic_signals.append({
                    "name": signal.name,
                    "start_bit": start_bit,
                    "bit_length": signal.size,
                    "factor": number_converter(signal.factor),
                    "offset": number_converter(signal.offset),
                    "is_big_endian": signal.is_little_endian is False,
                    "is_signed": signal.is_signed,
                    "is_float": signal.is_float,
                })
            symbolic_frame = {"name": frame.name,
                              "id": int(frame.arbitration_id.id),
                              "is_extended_frame": frame.arbitration_id.extended,
                              "is_fd": frame.is_fd,
                              "signals": symbolic_signals}
            frame_attributes = {
                attr: frame.attribute(attr)
                for attr in additional_frame_columns
                if frame.attribute(attr) is not None  # don't export None parameters
            }
            if frame_attributes:  # only add attributes if there are any
                symbolic_frame["attributes"] = frame_attributes
            export_dict['messages'].append(symbolic_frame)
    else:  # export_all
        _define_mapping = {"signal_defines": db.signal_defines, "frame_defines": db.frame_defines,
                           "global_defines": db.global_defines, "env_defines": db.env_defines, "ecu_defines": db.ecu_defines}
        for define_type in _define_mapping:
            export_dict[define_type] = [{"name": a,
                                         "define": _define_mapping[define_type][a].definition,
                                         "default": _define_mapping[define_type][a].defaultValue,
                                         "type": _define_mapping[define_type][a].type} for a in _define_mapping[define_type]]
        export_dict['ecus'] = {ecu.name: ecu.comment for ecu in db.ecus}
        export_dict['attributes'] = db.attributes
        export_dict['value_tables'] = db.value_tables
        export_dict['env_vars'] = db.env_vars
        export_dict['baudrate'] = db.baudrate
        export_dict['fd_baudrate'] = db.fd_baudrate

        for frame in db.frames:
            frame_attributes = {attribute: frame.attribute(attribute, db=db) for attribute in db.frame_defines}
            symbolic_signals = []
            for signal in frame.signals:
                attributes = {attribute: signal.attribute(attribute, db=db) for attribute in db.signal_defines}
                values = {key: signal.values[key] for key in signal.values}
                if not signal.is_little_endian:
                    if motorola_bit_format == "msb":
                        start_bit = signal.get_startbit(bit_numbering=1)
                    elif motorola_bit_format == "msbreverse":
                        start_bit = signal.get_startbit()
                    else:  # motorola_bit_format == "lsb"
                        start_bit = signal.get_startbit(bit_numbering=1, start_little=True)
                else:  # motorola_bit_format == "lsb"
                    start_bit = signal.get_startbit(bit_numbering=1, start_little=True)

                symbolic_signal = {
                    "name": signal.name,
                    "start_bit": start_bit,
                    "bit_length": signal.size,
                    "factor": number_converter(signal.factor),
                    "offset": number_converter(signal.offset),
                    "min": number_converter(signal.min),
                    "max": number_converter(signal.max),
                    "is_big_endian": signal.is_little_endian is False,
                    "is_signed": signal.is_signed,
                    "is_float": signal.is_float,
                    "comment": signal.comment,
                    "comments": signal.comments,
                    "attributes": attributes,
                    "initial_value": number_converter(signal.initial_value),
                    "values": values,
                    "is_multiplexer": signal.is_multiplexer,
                    "mux_value": signal.mux_val,
                    "receivers": signal.receivers,
                }
                if signal.multiplex is not None:
                    symbolic_signal["multiplex"] = signal.multiplex
                if signal.unit:
                    symbolic_signal["unit"] = signal.unit
                if signal.muxer_for_signal is not None:
                    symbolic_signal["muxer_for_signal"] = signal.muxer_for_signal
                if signal.mux_val_grp:
                    symbolic_signal["mux_val_grp"] = signal.mux_val_grp

                symbolic_signals.append(symbolic_signal)

            export_dict['messages'].append(
                {"name": frame.name,
                 "id": int(frame.arbitration_id.id),
                 "is_extended_frame": frame.arbitration_id.extended,
                 "is_fd": frame.is_fd,
                 "signals": symbolic_signals,
                 "attributes": frame_attributes,
                 "comment": frame.comment,
                 "length": frame.size,
                 "is_complex_multiplexed": frame.is_complex_multiplexed,
                 "mux_names": frame.mux_names,
                 "cycle_time": frame.cycle_time,
                 "is_j1939": frame.is_j1939,
                 "header_id": frame.header_id,
                 "pdu_name": frame.pdu_name,
                 "transmitters": frame.transmitters})
    if sys.version_info > (3, 0):
        import io
        temp = io.TextIOWrapper(f, encoding='UTF-8')
    else:
        temp = f

    try:
        json.dump(export_dict, temp, sort_keys=True,
                  indent=4, separators=(',', ': '))
    finally:
        if sys.version_info > (3, 0):
            # When TextIOWrapper is garbage collected, it closes the raw stream
            # unless the raw stream is detached first
            temp.detach()


def load(f, **_options):
    # type: (typing.BinaryIO, **str) -> canmatrix.CanMatrix

    db = canmatrix.CanMatrix()

    if sys.version_info > (3, 0):
        import io
        json_data = json.load(io.TextIOWrapper(f, encoding='UTF-8'))
    else:

        json_data = json.load(f)

    if "enumerations" in json_data:
        for val_tab_name, val_tab_dict in json_data['enumerations'].items():
            for key, val in val_tab_dict.items():
                if key.isdigit():
                    key = int(key)
                db.value_tables.setdefault(val_tab_name, {})[key] = val

    if "ecus" in json_data:
        for ecu in json_data["ecus"]:
            new_ecu = canmatrix.Ecu(name=ecu, comment=json_data["ecus"][ecu])
            db.add_ecu(new_ecu)
    if "messages" in json_data:
        for frame in json_data["messages"]:
            # new_frame = Frame(frame["id"],frame["name"],8,None)
            arb_id = canmatrix.canmatrix.ArbitrationId(id=frame["id"], extended=frame.get("is_extended_frame", "False"))
            new_frame = canmatrix.Frame(frame["name"], arbitration_id=arb_id, size=8)
            if "length" in frame:
                new_frame.size = frame["length"]
            simple_mapping = ["is_complex_multiplexed", "mux_names", "cycle_time", "is_j1939", "header_id", "pdu_name"]
            for key in simple_mapping:
                if key in frame:
                    if key == "is_complex_multiplexed":
                        new_frame.is_complex_multiplexed = frame[key]
                    elif key == "mux_names":
                        new_frame.mux_names = frame[key]
                    elif key == "cycle_time":
                        new_frame.cycle_time = frame[key]
                    elif key == "is_j1939":
                        new_frame.is_j1939 = frame[key]
                    elif key == "header_id":
                        new_frame.header_id = frame[key]
                    elif key == "pdu_name":
                        new_frame.pdu_name = frame[key]

            new_frame.arbitration_id.extended = frame.get("is_extended_frame", False)
            if "transmitters" in frame:
                new_frame.transmitters = frame["transmitters"]
            for signal in frame["signals"]:
                is_little_endian = not signal.get("is_big_endian", False)
                is_float = signal.get("is_float", False)
                is_signed = signal.get("is_signed", False)

                new_signal = canmatrix.Signal(
                    signal["name"],
                    start_bit=signal["start_bit"],
                    size=signal["bit_length"],
                    is_little_endian=is_little_endian,
                    is_signed=is_signed,
                    is_float=is_float,
                    factor=signal.get("factor", 1),
                    offset=signal.get("offset", 0)
                )

                if signal.get("min") is not None:
                    new_signal.min = new_signal.float_factory(signal["min"])

                if signal.get("max", False):
                    new_signal.max = new_signal.float_factory(signal["max"])

                if signal.get("unit", False):
                    new_signal.unit = signal["unit"]

                if signal.get("multiplex", False):
                    new_signal.multiplex = signal["multiplex"]

                if signal.get("values", False):
                    for key in signal["values"]:
                        new_signal.add_values(key, signal["values"][key])

                if signal.get("attributes", False):
                    for key in signal["attributes"]:
                        new_signal.add_attribute(key, signal["attributes"][key])

                if "comment" in signal:
                    new_signal.comment = signal["comment"]

                if "comments" in signal:
                    new_signal.comments = signal["comments"]

                if "initial_value" in signal:
                    new_signal.initial_value = decimal.Decimal(signal["initial_value"])

                if signal.get("receivers", False):
                    for ecu in signal["receivers"]:
                        new_signal.add_receiver(ecu)
                if new_signal.is_little_endian is False:
                    new_signal.set_startbit(
                        new_signal.start_bit, bitNumbering=1, startLittle=True)
                new_frame.add_signal(new_signal)
            db.add_frame(new_frame)

    _define_list = {"signal_defines": db.add_signal_defines, "frame_defines": db.add_frame_defines,
                    "global_defines": db.add_global_defines, "env_defines": db.add_env_defines,
                    "ecu_defines": db.add_ecu_defines}
    for define_type, fptr in _define_list.items():
        if define_type in json_data:
            for define in json_data[define_type]:
                fptr(define['name'], define['define'])

    cm_import_list_dict = {'attributes': db.attributes, 'value_tables': db.value_tables, 'env_vars': db.env_vars}

    cm_import_list_val = ['baudrate', 'fd_baudrate']

    for key in cm_import_list_dict:
        if key in json_data:
            cm_import_list_dict[key].update(json_data[key])

    for key in cm_import_list_val:
        if key in json_data:
            if key == 'baudrate':
                db.baudrate = json_data[key]
            elif key == 'fd_baudrate':
                db.fd_baudrate = json_data[key]

    f.close()
    db.update_ecu_list()
    return db



# -*- coding: utf-8 -*-
# Copyright (c) 2013, Eduard Broecker
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that
# the following conditions are met:
#
#    Redistributions of source code must retain the above copyright notice, this list of conditions and the
#    following disclaimer.
#    Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

#
# this script exports fibex-files from a canmatrix-object
# fibex-files are the network-matrix-definitions; Canmatrix exports CAN
# only (fibex: Field Bus Exchange Format //
# https://de.wikipedia.org/wiki/Field_Bus_Exchange_Format)

from __future__ import absolute_import, division, print_function

import typing
from builtins import *
import lxml.etree
import logging
import canmatrix
import re
import decimal

clusterImporter = 1

logger = logging.getLogger(__name__)

fx = "http://www.asam.net/xml/fbx"
ho = "http://www.asam.net/xml"
can = "http://www.asam.net/xml/fbx/can"
xsi = "http://www.w3.org/2001/XMLSchema-instance"
ns_ho = "{%s}" % ho
ns_fx = "{%s}" % fx
ns_can = "{%s}" % can
ns_xsi = "{%s}" % xsi

extension = "xml"

# noinspection PyProtectedMember
_Element = lxml.etree._Element


def create_short_name_desc(parent, short_name, desc):
    # type: (_Element, str, str) -> None
    short_name_elem = lxml.etree.SubElement(parent, ns_ho + "SHORT-NAME")
    short_name_elem.text = short_name
    desc_elem = lxml.etree.SubElement(parent, ns_ho + "DESC")
    desc_elem.text = desc


def create_sub_element_fx(parent, element_name, element_text=None):
    # type: (_Element, str, typing.Optional[str]) -> _Element
    new = lxml.etree.SubElement(parent, ns_fx + element_name)
    if element_text is not None:
        new.text = element_text
    return new


def create_sub_element_ho(parent, element_name, element_text=None):
    # type: (_Element, str, typing.Optional[str]) -> _Element
    new = lxml.etree.SubElement(parent, ns_ho + element_name)
    if element_text is not None:
        new.text = element_text
    return new


class Fe:
    def __init__(self, filename):
        self.tree = lxml.etree.parse(filename)
        self.root = self.tree.getroot()  # type: _Element

        self.ns = "{" + self.tree.xpath('namespace-uri(.)') + "}"  # type: str
        self.nsp = self.tree.xpath('namespace-uri(.)')
        self.ans = "{http://www.asam.net/xml}"
        self._id_cache = {a.attrib["ID"]:a for a in self.root.xpath(".//*[@ID]")}
        self._id_rev_cache = {}
        for referencer in self.root.xpath(".//*[@ID-REF]"):
            ref_id = referencer.attrib["ID-REF"]
            if ref_id not in self._id_rev_cache:
                self._id_rev_cache[ref_id] = [referencer]
            else:
                self._id_rev_cache[ref_id].append(referencer)

    def sn(self, tag):
        sn = tag.find("./" + self.ans + "SHORT-NAME").text
        return sn

    def get_referencable_parent(self, xml_element):
        path = ""
        while xml_element != self.root:
            try:
               current_short_name = self.sn(xml_element)
               return xml_element
            except AttributeError:
                pass
            xml_element = xml_element.getparent()
        return xml_element

    def find_parent(self, start_element, parent_tag):
        while start_element != self.root:
            if start_element.tag == self.ns + parent_tag or start_element == self.ans + parent_tag:
                return start_element
            start_element = start_element.getparent()
        return None

    def get_desc_or_longname(self, element):
        long_name_elements = self.selector(element, "./!LONG-NAME")
        if len(long_name_elements) > 0:
            return long_name_elements[0].text

        desc_elements = self.selector(element, "./!DESC")
        if len(desc_elements) > 0:
            return desc_elements[0].text

        return ""

    def selector(self, start_element, selector):
        start_pos = 0
        token = ""
        result_list = [start_element]
        last_found_token = 0
        while start_pos < len(selector):
            token_match = re.search(r'//|/!|/|!|>>|>!|<<!|<<|\^|>|$', selector[start_pos:])
            found_token = token_match.span()
            if start_pos > 0:  # at least one Token found...
                value = selector[last_found_token:start_pos + found_token[0]]
                if token == "//":
                    result_list = [c for a in result_list for c in a.findall(".//" + self.ns + value)]
                elif token == "/!":
                    result_list = [c for a in result_list for c in a.findall(".//" + self.ans + value)]
                elif token == "/":
                    if value == "..":
                        result_list = [a.getparent() for a in result_list]
                    else:
                        result_list = [a.find("./" + self.ns + value) for a in result_list]
                elif token == "!":
                    result_list = [a.find("./" + self.ans + value) for a in result_list]
                elif token == ">":
                    start_points = [a.find("./" + self.ns + value).attrib["ID-REF"] for a in result_list]
                    result_list = [self._id_cache[a] for a in start_points]
                elif token == "<<":
                    id_list = [a.attrib["ID"] for a in result_list]
                    result_list = [b for a in id_list for b in self._id_rev_cache[a]]
                    result_list = [a for a in result_list if self.get_referencable_parent(a) is not None and self.get_referencable_parent(a).tag.endswith(value)]
                elif token == "<<!":
                    id_list = [a.attrib["ID"] for a in result_list]
                    result_list = [b for a in id_list for b in self._id_rev_cache[a]]
                elif token == ">>":
                    start_points = [c.attrib["ID-REF"] for a in result_list for c in a.findall(".//" + self.ns + value)]
                    result_list = [self._id_cache[a] for a in start_points]
                elif token == ">!":
                    start_points = [c.attrib["ID-REF"] for a in result_list for c in a.findall(".//" + self.ans + value)]
                    result_list = [self._id_cache[a] for a in start_points]
                elif token == ">>":
                    start_points = [c.attrib["ID-REF"] for a in result_list for c in a.findall(".//" + self.ns + value)]
                    result_list = [self._id_cache[a] for a in start_points]
                elif token == "^":
                    result_list = [self.find_parent(a, value) for a in result_list if a is not None and self.find_parent(a, value) is not None]

            result_list = [a for a in result_list if a is not None]
            last_found_token = found_token[1] + start_pos
            token = selector[start_pos + found_token[0]:start_pos + found_token[1]]
            start_pos += found_token[1]
        return sorted(result_list, key=lambda element: element.sourceline)


def get_signals_for_pdu(fe, pdu, overall_startbit = 0):
    signals = []
    ecus = []
    for signal_instance in fe.selector(pdu, "//SIGNAL-INSTANCE"):
        byte_order_element = fe.selector(signal_instance, "/IS-HIGH-LOW-BYTE-ORDER")
        if byte_order_element[0].text == "false":
            is_little_endian = True
        else:
            is_little_endian = False

        start_bit = int(fe.selector(signal_instance, "/BIT-POSITION")[0].text, 0) + overall_startbit
        signal = fe.selector(signal_instance, ">SIGNAL-REF")[0]
        ecu_instance_refs = fe.selector(signal_instance, "<<ECU")
        receiver_ecus = []
        for ecu_instance_ref in ecu_instance_refs:
            if len(fe.selector(ecu_instance_ref, "^INPUT-PORT")) > 0:
                ecu_name = fe.sn(fe.get_referencable_parent(ecu_instance_ref))
                receiver_ecus.append(ecu_name)
                ecus.append(canmatrix.Ecu(name=ecu_name.strip()))

        signal_name = fe.sn(signal)
        coding = fe.selector(signal, ">CODING-REF")[0]
        bit_length = int(fe.selector(coding, "/!BIT-LENGTH")[0].text)
        compu_methods = fe.selector(coding, "/!COMPU-METHOD")
        sig = canmatrix.Signal(name=signal_name)

        for compu_method in compu_methods:
            category = fe.selector(compu_method, "/!CATEGORY")
            if len(category) > 0 and category[0].text == "LINEAR":
                numerator = fe.selector(compu_method, "/!COMPU-NUMERATOR")[0]
                denominator = fe.selector(compu_method, "/!COMPU-DENOMINATOR")[0]
                teiler = decimal.Decimal(fe.selector(denominator, "/!V")[0].text)
                [offset, factor] = [decimal.Decimal(a.text) for a in fe.selector(numerator, "/!V")]
                [offset, factor] = [a / teiler for a in [offset, factor]]
                sig.offset = offset
                sig.factor = factor
                try:
                    sig.min = decimal.Decimal(fe.selector(compu_method, "!PHYS-CONSTRS!LOWER-LIMIT")[0].text)
                    sig.max = decimal.Decimal(fe.selector(compu_method, "!PHYS-CONSTRS!UPPER-LIMIT")[0].text)
                except:
                    pass
                unit = fe.selector(compu_method, ">!UNIT-REF")
                if len(unit) > 0:
                    try:
                        sig.unit = fe.selector(unit[0], "!DISPLAY-NAME")[0].text
                    except:
                        pass
            elif len(category) > 0 and category[0].text == "TEXTTABLE":
                for compu_scale in fe.selector(compu_method, "/!COMPU-SCALE"):
                    try:
                        value_name = fe.selector(compu_scale, "!COMPU-CONST!VT")[0].text
                    except IndexError:
                        value_name = fe.get_desc_or_longname(compu_scale)
                    value_value = fe.selector(compu_scale, "!LOWER-LIMIT")[0].text
                    sig.add_values(value_value, value_name)
        sig.is_little_endian = is_little_endian
        if not sig.is_little_endian:
            sig.set_startbit(start_bit, bitNumbering=1)
        else:
            sig.start_bit = start_bit
        sig.size = bit_length
        sig.receivers = list(set(receiver_ecus))

        sig.add_comment(fe.get_desc_or_longname(signal))
        signals.append(sig)
    return signals, ecus


def load(f, **_options):
    fe = Fe(f)
    result = {}
    sig_group_counter = 0

    clusters = fe.selector(fe.root, "//CLUSTER")
    names = [fe.sn(a) for a in clusters]
    logger.info("Found clusters: " + ",".join(names))

    for cluster in clusters:
        if "CAN" not in fe.selector(cluster, "//PROTOCOL")[0].text:
            logger.info(fe.sn(cluster) + " seems not to be a CAN cluster - ignoring")
            continue

        db = canmatrix.CanMatrix()
        result[fe.sn(cluster)] = db
        channels = fe.selector(cluster, ">>CHANNEL-REF")
        for channel in channels:
            for ft in fe.selector(channel, "//FRAME-TRIGGERING"):
                ports = fe.selector(ft, "<<ECU/..")
                output_ports = [a for a in ports if a.tag == fe.ns + "OUTPUT-PORT"]
                sending_ecus = [fe.get_referencable_parent(a) for a in output_ports]
                arbitration_id_element = fe.selector(ft, "//IDENTIFIER-VALUE")[0]
                arbitration_id = int(arbitration_id_element.text)
                extended = arbitration_id_element.attrib.get("EXTENDED-ADDRESSING", 'false') == 'true'
                frame_element = fe.selector(ft, ">FRAME-REF")[0]
                frame_size = int(fe.selector(frame_element, "/BYTE-LENGTH")[0].text)

                pdu_instances = fe.selector(frame_element, "//PDU-INSTANCE")

                if len(pdu_instances) > 1:
                    frame_name = fe.sn(frame_element)
                    frame = canmatrix.Frame(name=frame_name)
                    for pdu_instance in pdu_instances:
                        pdu = fe.selector(pdu_instance, ">PDU-REF")[0]
                        pdu_startbit_position = int(fe.selector(pdu_instance, "/BIT-POSITION")[0].text, 0)
                        signals, ecus = get_signals_for_pdu(fe, pdu, pdu_startbit_position)
                        for sig in signals:
                            frame.add_signal(sig)
                        for ecu in ecus:
                            db.add_ecu(ecu)

                        frame.add_signal_group(fe.sn(pdu), sig_group_counter, [sig.name for sig in signals])
                        sig_group_counter += 1
                else:
                    pdu = fe.selector(pdu_instances[0], ">PDU-REF")[0]
                    frame_name = fe.sn(pdu)
                    frame = canmatrix.Frame(name=frame_name)

                    signals, ecus = get_signals_for_pdu(fe, pdu)
                    for sig in signals:
                        frame.add_signal(sig)
                    for ecu in ecus:
                        db.add_ecu(ecu)

                #                fe.selector(pdu, "<<!PDU-TRIGGERING")
                frame.size = frame_size
                if len(output_ports) > 0:
                    pdu_triggerings = fe.selector(output_ports[0], ">>PDU-TRIGGERING-REF")
                    if len(pdu_triggerings) > 0:
                        cyclic_timing_element = fe.selector(pdu_triggerings[0],
                                                    "/TIMINGS/CYCLIC-TIMING/REPEATING-TIME-RANGE/VALUE")
                        if len(cyclic_timing_element) > 0:
                            time_value_string = cyclic_timing_element[0].text
                            if time_value_string.startswith("PT") and time_value_string.endswith("S"):
                                frame.cycle_time = decimal.Decimal(time_value_string[2:-1])*1000
                frame.transmitters = [fe.sn(a) for a in sending_ecus]
                for ecu_element in sending_ecus:
                    ecu_name = fe.sn(ecu_element)
                    cm_ecu = canmatrix.Ecu(ecu_name)
                    cm_ecu.add_comment(fe.get_desc_or_longname(ecu_element))
                    db.add_ecu(cm_ecu)
                frame.arbitration_id = canmatrix.ArbitrationId(extended=extended, id=arbitration_id)

                frame.add_comment(fe.get_desc_or_longname(pdu))
                if "CAN-FD" in [a.text for a in
                     fe.selector(ft, "//CAN-FRAME-TX-BEHAVIOR") + fe.selector(ft, "//CAN-FRAME-RX-BEHAVIOR")]:
                    frame.is_fd = True

                db.add_frame(frame)
    return result


def dump(db, f, **options):
    # type: (canmatrix.CanMatrix, typing.IO, **typing.Any) -> None
    ns_map = {"fx": fx, "ho": ho, "can": can, "xsi": xsi}
    root = lxml.etree.Element(ns_fx + "FIBEX", nsmap=ns_map)
    root.attrib[
        '{{{pre}}}schemaLocation'.format(
            pre=xsi)] = 'http://www.asam.net/xml/fbx ..\\..\\xml_schema\\fibex.xsd http://www.asam.net/xml/fbx/can  ..\\..\\xml_schema\\fibex4can.xsd'

    #
    # PROJECT
    #
    project = create_sub_element_fx(root, "PROJECT")
    project.set('ID', 'canmatrixExport')
    create_short_name_desc(project, "CAN", "Canmatrix Export")

    #
    # ELEMENTS
    #
    elements = create_sub_element_fx(root, "ELEMENTS")
    
    #
    # CLUSTERS
    #
    clusters = create_sub_element_fx(elements, "CLUSTERS")
    cluster = lxml.etree.SubElement(clusters, ns_fx + "CLUSTER")
    cluster.set('ID', 'canCluster1')
    create_short_name_desc(cluster, "clusterShort", "clusterDesc")
    create_sub_element_fx(cluster, "SPEED", "500")
    create_sub_element_fx(cluster, "IS-HIGH-LOW-BIT-ORDER", "true")
    create_sub_element_fx(cluster, "BIT-COUNTING-POLICY", "MONOTONE")
    protocol = create_sub_element_fx(cluster, "PROTOCOL", "CAN")
    protocol.attrib['{{{pre}}}type'.format(pre=xsi)] = "can:PROTOCOL-TYPE"
    create_sub_element_fx(cluster, "PROTOCOL-VERSION", "20")
    channel_refs = create_sub_element_fx(cluster, "CHANNEL-REFS")
    # for each channel
    channel_ref = create_sub_element_fx(channel_refs, "CHANNEL-REF")
    channel_ref.set("ID-REF", "CANCHANNEL01")

    #
    # CHANNELS
    #
    channels = create_sub_element_fx(elements, "CHANNELS")
    channel = create_sub_element_fx(channels, "CHANNEL")
    # for each channel
    channel.set('ID', 'CANCHANNEL01')
    create_short_name_desc(channel, "CANCHANNEL01", "Can Channel Description")

    # for pdu triggerings
    pdu_triggerings = create_sub_element_fx(channel, "PDU-TRIGGERINGS")
    for pdu in db.frames:
        pdu_triggering = create_sub_element_fx(
            pdu_triggerings, "PDU-TRIGGERING")
        pdu_triggering.set("ID", "PDU_" + pdu.name)
        pdu_timings = create_sub_element_fx(pdu_triggering, "TIMINGS")
        if pdu.cycle_time > 0:
            cyclic_timing = create_sub_element_fx(pdu_timings, "CYCLIC-TIMING")
            repeating_time_range = create_sub_element_fx(cyclic_timing, "REPEATING-TIME-RANGE")
            create_sub_element_fx(repeating_time_range, "VALUE", "PT" + str(pdu.cycle_time/1000.0) + "S")

        pdu_ref = create_sub_element_fx(pdu_triggering, "PDU-REF")
        pdu_ref.set("ID-REF", "PDU_" + pdu.name)

    frame_triggerings = create_sub_element_fx(channel, "FRAME-TRIGGERINGS")
    for frame in db.frames:
        frame_triggering = create_sub_element_fx(
            frame_triggerings, "FRAME-TRIGGERING")
        frame_triggering.set("ID", "FT_" + frame.name)
        identifier = create_sub_element_fx(frame_triggering, "IDENTIFIER")
        create_sub_element_fx(identifier, "IDENTIFIER-VALUE", str(frame.arbitration_id.id))
        frame_ref = create_sub_element_fx(frame_triggering, "FRAME-REF")
        frame_ref.set("ID-REF", "FRAME_" + frame.name)

    #
    # ECUS
    #
    ecus = create_sub_element_fx(elements, "ECUS")
    for bu in db.ecus:
        ecu = create_sub_element_fx(ecus, "ECU")
        ecu.set("ID", bu.name)
        create_short_name_desc(ecu, bu.name, bu.comment)
        function_refs = create_sub_element_fx(ecu, "FUNCTION-REFS")
        func_ref = create_sub_element_fx(function_refs, "FUNCTION-REF")
        func_ref.set("ID-REF", "FCT_" + bu.name)

        controllers = create_sub_element_fx(ecu, "CONTROLLERS")
        controller = create_sub_element_fx(controllers, "CONTROLLER")
        create_short_name_desc(controller, bu.name, bu.comment)
        controller.set('ID', 'Controller_' + bu.name)

        connectors = create_sub_element_fx(ecu, "CONNECTORS")
        connector = create_sub_element_fx(connectors, "CONNECTOR")
        connector.set('ID', 'Connector' + bu.name)
        channel_ref = create_sub_element_fx(connector, "CHANNEL-REF")
        channel_ref.set("ID-REF", "CANCHANNEL01")
        controller_ref = create_sub_element_fx(connector, "CONTROLLER-REF")
        controller_ref.set("ID-REF", 'Controller_' + bu.name)
        inputs = create_sub_element_fx(connector, "INPUTS")
        for frame in db.frames:
            if bu.name in frame.receivers:
                input_port = create_sub_element_fx(inputs, "INPUT-PORT")
                frame_triggering_ref = create_sub_element_fx(input_port, "FRAME-TRIGGERING-REF")
                frame_triggering_ref.set("ID-REF", "FT_" + frame.name)
                # Reference to PDUs
                included_pdus = create_sub_element_fx(input_port, "INCLUDED-PDUS")
                included_pdu = create_sub_element_fx(included_pdus, "INCLUDED-PDU")
                included_pdu.set('ID', 'input_included_pdu_' + frame.name)
                pdu_triggering_ref = create_sub_element_fx(included_pdu, "PDU-TRIGGERING-REF")
                pdu_triggering_ref.set("ID-REF", "PDU_" + frame.name)

                
        outputs = create_sub_element_fx(connector, "OUTPUTS")
        for frame in db.frames:
            if bu.name in frame.transmitters:
                input_port = create_sub_element_fx(outputs, "OUTPUT-PORT")
                frame_triggering_ref = create_sub_element_fx(input_port, "FRAME-TRIGGERING-REF")
                frame_triggering_ref.set("ID-REF", "FT_" + frame.name)
                # Reference to PDUs
                included_pdus = create_sub_element_fx(input_port, "INCLUDED-PDUS")
                included_pdu = create_sub_element_fx(included_pdus, "INCLUDED-PDU")
                included_pdu.set('ID', 'output_included_pdu_' + frame.name)
                pdu_triggering_ref = create_sub_element_fx(included_pdu, "PDU-TRIGGERING-REF")
                pdu_triggering_ref.set("ID-REF", "PDU_" + frame.name)
                
        
        # ignore CONTROLERS/CONTROLER

    #
    # PDUS
    #
    pdus = create_sub_element_fx(elements, "PDUS")
    for frame in db.frames:
        pdu = create_sub_element_fx(pdus, "PDU")
        pdu.set("ID", "PDU_" + frame.name)
        create_short_name_desc(pdu, "PDU_" + frame.name, frame.comment)
        create_sub_element_fx(pdu, "BYTE-LENGTH", str(frame.size))  # DLC
        create_sub_element_fx(pdu, "PDU-TYPE", "APPLICATION")
        signal_instances = create_sub_element_fx(pdu, "SIGNAL-INSTANCES")
        for signal in frame.signals:
            signal_id = frame.name + "." + signal.name
            signal_instance = create_sub_element_fx(
                signal_instances, "SIGNAL-INSTANCE")
            signal_instance.set("ID", "PDUINST_" + signal_id)
            # startBit: TODO - find out correct BYTEORDER ...
            create_sub_element_fx(signal_instance, "BIT-POSITION",
                                  str(signal.start_bit))
            if signal.is_little_endian:
                create_sub_element_fx(
                    signal_instance,
                    "IS-HIGH-LOW-BYTE-ORDER",
                    "false")  # true:big endian; false:little endian
            else:
                create_sub_element_fx(
                    signal_instance, "IS-HIGH-LOW-BYTE-ORDER", "true")
            signal_ref = create_sub_element_fx(signal_instance, "SIGNAL-REF")
            signal_ref.set("ID-REF", "SIG_" + signal_id)

    # FRAMES
    #
    frames = create_sub_element_fx(elements, "FRAMES")
    for frame in db.frames:
        frame_element = create_sub_element_fx(frames, "FRAME")
        frame_element.set("ID", "FRAME_" + frame.name)
        create_short_name_desc(frame_element, frame.name, frame.comment)
        create_sub_element_fx(frame_element, "BYTE-LENGTH", str(frame.size))  # DLC
        create_sub_element_fx(frame_element, "PDU-TYPE", "APPLICATION")
        pdu_instances = create_sub_element_fx(frame_element, "PDU-INSTANCES")
        pdu_instance = create_sub_element_fx(pdu_instances, "PDU-INSTANCE")
        pdu_instance.set("ID", "PDUINSTANCE_" + frame.name)
        pdu_ref = create_sub_element_fx(pdu_instance, "PDU-REF")
        pdu_ref.set("ID-REF", "PDU_" + frame.name)
        create_sub_element_fx(pdu_instance, "BIT-POSITION", "0")
        create_sub_element_fx(pdu_instance, "IS-HIGH-LOW-BYTE-ORDER", "false")

    #
    # FUNCTIONS
    #
    functions = create_sub_element_fx(elements, "FUNCTIONS")
    for bu in db.ecus:
        function = create_sub_element_fx(functions, "FUNCTION")
        function.set("ID", "FCT_" + bu.name)
        create_short_name_desc(function, "FCT_" + bu.name, bu.comment)
        input_ports = create_sub_element_fx(function, "INPUT-PORTS")
        for frame in db.frames:
            for signal in frame.signals:
                signal_id = frame.name + "." + signal.name
                if bu.name in signal.receivers:
                    input_port = create_sub_element_fx(input_ports, "INPUT-PORT")
                    input_port.set("ID", "INP_" + signal_id)
                    desc = lxml.etree.SubElement(input_port, ns_ho + "DESC")
                    desc.text = signal.comment
                    signal_ref = create_sub_element_fx(input_port, "SIGNAL-REF")
                    signal_ref.set("ID-REF", "SIG_" + signal_id)

        output_ports = create_sub_element_fx(function, "OUTPUT-PORTS")
        for frame in db.frames:
            if bu.name in frame.transmitters:
                for signal in frame.signals:
                    signal_id = frame.name + "." + signal.name
                    output_port = create_sub_element_fx(output_ports, "OUTPUT-PORT")
                    output_port.set("ID", "OUTP_" + signal_id)
                    desc = lxml.etree.SubElement(output_port, ns_ho + "DESC")
                    desc.text = "signalcomment"
                    signal_ref = create_sub_element_fx(output_port, "SIGNAL-REF")
                    signal_ref.set("ID-REF", "SIG_" + frame.name + "_" + signal_id)

    #
    # SIGNALS
    #
    signals = create_sub_element_fx(elements, "SIGNALS")
    for frame in db.frames:
        for signal in frame.signals:
            signal_id = frame.name + "." + signal.name
            signal_element = create_sub_element_fx(signals, "SIGNAL")
            signal_element.set("ID", "SIG_" + signal_id)
            create_short_name_desc(signal_element, signal.name, signal.comment)
            coding_ref = create_sub_element_fx(signal_element, "CODING-REF")
            coding_ref.set("ID-REF", "CODING_" + signal_id)

    #
    # PROCESSING-INFORMATION
    #
    proc_info = lxml.etree.SubElement(root, ns_fx + "PROCESSING-INFORMATION", nsmap={"ho": ho})
    unit_spec = create_sub_element_ho(proc_info, "UNIT-SPEC")
    for frame in db.frames:
        for signal in frame.signals:
            signal_id = frame.name + "." + signal.name
            unit = create_sub_element_ho(unit_spec, "UNIT")
            unit.set("ID", "UNIT_" + signal_id)
            create_sub_element_ho(unit, "SHORT-NAME", signal.name)
            create_sub_element_ho(unit, "DISPLAY-NAME", signal.unit)

    codings = create_sub_element_fx(proc_info, "CODINGS")
    for frame in db.frames:
        for signal in frame.signals:
            signal_id = frame.name + "." + signal.name
            coding = create_sub_element_fx(codings, "CODING")
            coding.set("ID", "CODING_" + signal_id)
            create_short_name_desc(
                coding,
                "CODING_" +
                signal_id,
                "Coding for " +
                signal_id)
                
            coded = create_sub_element_ho(coding, "CODED-TYPE")
            create_sub_element_ho(coded, "BIT-LENGTH", str(signal.size))
            
            compu_methods = create_sub_element_ho(coding, "COMPU-METHODS")
            compu_method = create_sub_element_ho(compu_methods, "COMPU-METHOD")
            create_sub_element_ho(
                compu_method,
                "SHORT-NAME",
                "COMPUMETHOD_" +
                signal.name)
            create_sub_element_ho(compu_method, "CATEGORY", "LINEAR")
            unit_ref = create_sub_element_ho(compu_method, "UNIT-REF")
            unit_ref.set("ID-REF", "UNIT_" + signal_id)
            compu_internal_to_phys = create_sub_element_ho(
                compu_method, "COMPU-INTERNAL-TO-PHYS")
            compu_scales = create_sub_element_ho(
                compu_internal_to_phys, "COMPU-SCALES")
            compu_scale = create_sub_element_ho(compu_scales, "COMPU-SCALE")
            lower_limit = create_sub_element_ho(compu_scale, "LOWER-LIMIT", str(signal.min))  # Min
            lower_limit.set("INTERVAL-TYPE", "CLOSED")
            upper_limit = create_sub_element_ho(compu_scale, "UPPER-LIMIT", str(signal.max))  # Max
            upper_limit.set("INTERVAL-TYPE", "CLOSED")

            compu_rational_coeffs = create_sub_element_ho(
                compu_scale, "COMPU-RATIONAL-COEFFS")
            compu_numerator = create_sub_element_ho(
                compu_rational_coeffs, "COMPU-NUMERATOR")
            create_sub_element_ho(compu_numerator, "V", str(signal.offset))  # offset
            create_sub_element_ho(compu_numerator, "V", str(signal.factor))  # factor
            compu_denominator = create_sub_element_ho(compu_rational_coeffs, "COMPU-DENOMINATOR")
            create_sub_element_ho(compu_denominator, "V", "1")  # nenner
            # defaultValue = create_sub_element_ho(compuInternalToPhys,"COMPU-DEFAULT-VALUE")

            # CAN signal interpretation
            if signal.values:
                compu_method =  create_sub_element_ho(compu_methods, "COMPU-METHOD")
                create_sub_element_ho(compu_method, "SHORT-NAME",
                    "SIGNAL_INTERPRETATION_" + signal.name)
                create_sub_element_ho(compu_method, "CATEGORY", "TEXTTABLE")
                compu_int_to_phys = create_sub_element_ho(compu_method, "COMPU-INTERNAL-TO-PHYS")
                compu_scales = create_sub_element_ho(compu_int_to_phys, "COMPU-SCALES")

                for value, text in signal.values.items():
                    compu_scale = create_sub_element_ho(compu_scales, "COMPU-SCALE")

                    lower_limit = create_sub_element_ho(compu_scale, "LOWER-LIMIT", str(value))
                    lower_limit.set("INTERVAL-TYPE", "CLOSED")
                    upper_limit = create_sub_element_ho(compu_scale, "UPPER-LIMIT", str(value))
                    upper_limit.set("INTERVAL-TYPE", "CLOSED")

                    compu_const = create_sub_element_ho(compu_scale, "COMPU-CONST")
                    create_sub_element_ho(compu_const, "VT", text)

    #
    # REQUIREMENTS
    #
    # requirements = createSubElementFx(elements,  "REQUIREMENTS")

    f.write(lxml.etree.tostring(root, pretty_print=True))



import canmatrix.formats.arxml
import textwrap
import io


def test_pdu_gateway():
    arxml = io.BytesIO(textwrap.dedent(u'''\
    <AUTOSAR xmlns="http://autosar.org/schema/r4.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://autosar.org/schema/r4.0 autosar_4-3-0.xsd">

    <AR-PACKAGE>
      <SHORT-NAME>Cluster</SHORT-NAME>
      <ELEMENTS>
        <CAN-CLUSTER>
          <SHORT-NAME>someCluster</SHORT-NAME>
          <CAN-CLUSTER-VARIANTS>
            <CAN-CLUSTER-CONDITIONAL>
              <PHYSICAL-CHANNELS>
                <CAN-PHYSICAL-CHANNEL>
                  <SHORT-NAME>CHNL</SHORT-NAME>
                   <FRAME-TRIGGERINGS>
                    <CAN-FRAME-TRIGGERING>
                      <SHORT-NAME>SomeFrameTriggering</SHORT-NAME>
                      <PDU-TRIGGERINGS>
                        <PDU-TRIGGERING-REF-CONDITIONAL>
                          <PDU-TRIGGERING-REF DEST="PDU-TRIGGERING">/Cluster/someCluster/CHNL/somePDUTriggering</PDU-TRIGGERING-REF>
                        </PDU-TRIGGERING-REF-CONDITIONAL>
                      </PDU-TRIGGERINGS>
                    </CAN-FRAME-TRIGGERING>
                   </FRAME-TRIGGERINGS>
                   <I-SIGNAL-TRIGGERINGS>
                    <I-SIGNAL-TRIGGERING>
                      <SHORT-NAME>someSignalTriggering</SHORT-NAME>
                      <I-SIGNAL-REF DEST="I-SIGNAL">/ISignal/someSignal</I-SIGNAL-REF>
                    </I-SIGNAL-TRIGGERING>
                   </I-SIGNAL-TRIGGERINGS>
                   <PDU-TRIGGERINGS>
                    <PDU-TRIGGERING>
                      <SHORT-NAME>somePduTriggering</SHORT-NAME>
                      <I-PDU-REF DEST="NM-PDU">/PDU/somePdu</I-PDU-REF>
                    </PDU-TRIGGERING>
                   </PDU-TRIGGERINGS>
                </CAN-PHYSICAL-CHANNEL>
              </PHYSICAL-CHANNELS>
            </CAN-CLUSTER-CONDITIONAL>
          </CAN-CLUSTER-VARIANTS>
        </CAN-CLUSTER>

        <CAN-CLUSTER>
          <SHORT-NAME>someOtherCluster</SHORT-NAME>
          <CAN-CLUSTER-VARIANTS>
            <CAN-CLUSTER-CONDITIONAL>
              <PHYSICAL-CHANNELS>
                <CAN-PHYSICAL-CHANNEL>
                  <SHORT-NAME>CHNL</SHORT-NAME>
                   <FRAME-TRIGGERINGS>
                    <CAN-FRAME-TRIGGERING>
                      <SHORT-NAME>SomeOtherFrameTriggering</SHORT-NAME>
                      <PDU-TRIGGERINGS>
                        <PDU-TRIGGERING-REF-CONDITIONAL>
                          <PDU-TRIGGERING-REF DEST="PDU-TRIGGERING">/Cluster/someCluster/CHNL/someOtherPDUTrigering</PDU-TRIGGERING-REF>
                        </PDU-TRIGGERING-REF-CONDITIONAL>
                      </PDU-TRIGGERINGS>
                    </CAN-FRAME-TRIGGERING>
                   </FRAME-TRIGGERINGS>
                   <I-SIGNAL-TRIGGERINGS>
                    <I-SIGNAL-TRIGGERING>
                      <SHORT-NAME>someOtherSignalTriggering</SHORT-NAME>
                      <I-SIGNAL-REF DEST="I-SIGNAL">/ISignal/someOtherSignal</I-SIGNAL-REF>
                    </I-SIGNAL-TRIGGERING>
                   </I-SIGNAL-TRIGGERINGS>
                   <PDU-TRIGGERINGS>
                    <PDU-TRIGGERING>
                      <SHORT-NAME>someOtherPduTriggering</SHORT-NAME>
                      <I-PDU-REF DEST="NM-PDU">/PDU/someOtherPdu</I-PDU-REF>
                    </PDU-TRIGGERING>
                   </PDU-TRIGGERINGS>
                </CAN-PHYSICAL-CHANNEL>
              </PHYSICAL-CHANNELS>
            </CAN-CLUSTER-CONDITIONAL>
          </CAN-CLUSTER-VARIANTS>
        </CAN-CLUSTER>


      </ELEMENTS>
    </AR-PACKAGE>
    <AR-PACKAGE>
        <SHORT-NAME>Gateway</SHORT-NAME>
        <ELEMENTS>
            <GATEWAY>
              <SHORT-NAME>someECU</SHORT-NAME>
              <ECU-REF DEST="ECU-INSTANCE">/ECU/someECU</ECU-REF>
              <I-PDU-MAPPINGS>
                <I-PDU-MAPPING>
                <SOURCE-I-PDU-REF DEST="PDU-TRIGGERING">/Cluster/someCluster/CHNL/somePduTriggering</SOURCE-I-PDU-REF>
                <TARGET-I-PDU>
                    <TARGET-I-PDU-REF DEST="PDU-TRIGGERING">/Cluster/someOtherCluster/CHNL/someOtherPduTriggering</TARGET-I-PDU-REF>
                </TARGET-I-PDU>
                </I-PDU-MAPPING>
              </I-PDU-MAPPINGS>
             <SIGNAL-MAPPINGS>
                <I-SIGNAL-MAPPING>
                  <SOURCE-SIGNAL-REF DEST="I-SIGNAL-TRIGGERING">/Cluster/someCluster/CHNL/someSignalTriggering</SOURCE-SIGNAL-REF>
                  <TARGET-SIGNAL-REF DEST="I-SIGNAL-TRIGGERING">/Cluster/someOtherCluster/CHNL/someOtherSignalTriggering</TARGET-SIGNAL-REF>
                </I-SIGNAL-MAPPING>
             </SIGNAL-MAPPINGS>
            </GATEWAY>
        </ELEMENTS>
    </AR-PACKAGE>
    </AUTOSAR>''').encode('utf-8'))

    cluster = canmatrix.formats.arxml.load(arxml)

    assert cluster.get_pdu_routing_info("someOtherPdu")["target"][0]["pdu"] == "/PDU/somePdu"
    assert cluster.get_pdu_routing_info("/PDU/somePdu", strict_search=True)["source"][0]["pdu"] == "/PDU/someOtherPdu"
    assert cluster.get_signal_routing_info("someSignal")["source"][0]["signal"] == '/ISignal/someOtherSignal'
    assert cluster.get_signal_routing_info("/ISignal/someOtherSignal")["target"][0]["signal"] == '/ISignal/someSignal'
